-- Filename    	: ControlSNAndPi.vhd
-- Author:	   	: Andreas Torno
-- Description 	: This component generates control signal for all CN-FUs

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.widths.all;
USE work.delays.all;

ENTITY ControlSNAndPi IS
    PORT (
		-- Global control ports
		clock							: IN  std_logic;
		enable							: IN  std_logic;
		reset							: IN  std_logic;
		-- Input control ports
		in_enableRom					: IN  std_logic;
		in_oddIteration					: IN  std_logic;
		in_variableDelay				: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_addressRom					: IN  std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
		-- Output control ports
		out_newSD_readEnable			: OUT std_logic;
		out_oldSD_readEnable			: OUT std_logic;
		out_oldSD_address_vector		: OUT std_logic_vector((ROM_SN_WIDTH - 2) * PARALLELISM - 1 downto 0);
		out_newSD_address_vector		: OUT std_logic_vector(ROM_SN_WIDTH * PARALLELISM - 1 downto 0);
		out_permutationIndices_vector	: OUT std_logic_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 0)
    );
END ControlSNAndPi;

ARCHITECTURE RTL OF ControlSNAndPi IS
	COMPONENT ShiftRegister
	GENERIC (
		size 				: integer;
		width 				: integer
	);
	PORT (
		clock				: IN  std_logic;
		enable				: IN  std_logic;
		reset				: IN  std_logic;
		input     			: IN  std_logic_vector(width - 1 downto 0);
		output    			: OUT std_logic_vector(width - 1 downto 0)
	);
	END COMPONENT;
	
	COMPONENT VariableShiftRegister
	GENERIC (
		size 				: integer;
		width 				: integer;
		offset				: integer
	);
    PORT (
		clock				: IN  std_logic;
		enable				: IN  std_logic;
		reset				: IN  std_logic;
		active_size 		: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
        input     			: IN  std_logic_vector(width - 1 downto 0);
        output    			: OUT std_logic_vector(width - 1 downto 0)
    );
	END COMPONENT;
	
	COMPONENT ROM
	GENERIC (
		size 				: integer;
		width_data 			: integer;
		width_address		: integer;
		filename			: string
	);
    PORT (
		clock				: IN  std_logic;
		enable				: IN  std_logic;
		reset				: IN  std_logic;
		r_enable			: IN  std_logic;
		address				: IN  std_logic_vector(width_address - 1 downto 0);
        output    			: OUT std_logic_vector(width_data - 1 downto 0)
    );
	END COMPONENT;
	
	signal romSN_output						: std_logic_vector(ROM_SN_WIDTH * PARALLELISM - 1 downto 0);
	signal romPi0_output					: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal romPi2_output					: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal romPi2_output_inverted			: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal romPi0_readEnable				: std_logic;
	signal romPi2_readEnable				: std_logic;
	signal romPi0_address					: std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
	signal romPi2_address					: std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
	signal romPi0_variableDelay				: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal romPi2_variableDelay				: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal oldSD_readEnable					: std_logic;
	
	signal permutationIndicesPi0			: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal permutationIndicesPi1			: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal permutationIndicesPi1_inverted	: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal permutationIndicesPi1_output		: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal permutationIndicesPi2			: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
	signal permutationIndicesPi3			: std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);	
BEGIN
	
	-- Instantiation of Components for SN control signals
	RomSN : ROM
	GENERIC MAP(
		size 				=> ROM_SIZE,
		width_data 			=> ROM_SN_WIDTH * PARALLELISM,
		width_address		=> ROM_ADDRESS_WIDTH,
		filename			=> "RomSn"
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		r_enable			=> in_enableRom,
		address				=> in_addressRom,
        output    			=> romSN_output
    );
	
	delay_RomSN : VariableShiftRegister
	GENERIC MAP(
		size 				=> MAX_DEGREE,
		width 				=> ROM_SN_WIDTH * PARALLELISM,
		offset				=> DELAY_NEW_SD
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		active_size 		=> in_variableDelay,
        input     			=> romSN_output,
        output    			=> out_newSD_address_vector
    );
	
	delay_readEnable : ShiftRegister
	GENERIC MAP(
		size 				=> DURATION_MEM_ACCESS,
		width 				=> 1
	)
	PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		input(0)     		=> in_enableRom,
		output(0)  			=> oldSD_readEnable
	);
	out_oldSD_readEnable	<= oldSD_readEnable;
	
	delay_newSD_readEnable : VariableShiftRegister
	GENERIC MAP(
		size 				=> MAX_DEGREE,
		width 				=> 1,
		offset				=> DELAY_NEW_SD
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		active_size 		=> in_variableDelay,
        input(0)   			=> oldSD_readEnable,
        output(0)  			=> out_newSD_readEnable
    );
	
	Output_oldSD_address : process(romSN_output)
	variable max : integer range 0 to PARALLELISM * (ROM_SN_WIDTH - 2) - 1;
	variable min : integer range 0 to PARALLELISM * (ROM_SN_WIDTH - 2) - 1;
	BEGIN
		for i in 0 to PARALLELISM - 1 loop
			max := (PARALLELISM - i) * (ROM_SN_WIDTH - 2) - 1;
			min := (PARALLELISM - i - 1) * (ROM_SN_WIDTH - 2);
			out_oldSD_address_vector(max downto min) <= romSN_output((PARALLELISM - i)*ROM_SN_WIDTH - 1 downto (PARALLELISM - i - 1)*ROM_SN_WIDTH + 2);
		end loop;
	END process;
	
	-- Instantiation of Components for permutation networks control signals
	RomPi0 : ROM
	GENERIC MAP(
		size 				=> ROM_SIZE,
		width_data 			=> ROM_PERMUTATION_WIDTH,
		width_address		=> ROM_ADDRESS_WIDTH,
		filename			=> "RomPi0"
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		r_enable			=> romPi0_readEnable,
		address				=> romPi0_address,
        output    			=> romPi0_output
    );
	permutationIndicesPi0	<= romPi0_output;
	
	RomPi2 : ROM
	GENERIC MAP(
		size 				=> ROM_SIZE,
		width_data 			=> ROM_PERMUTATION_WIDTH,
		width_address		=> ROM_ADDRESS_WIDTH,
		filename			=> "RomPi2"
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		r_enable			=> romPi2_readEnable,
		address				=> romPi2_address,
        output    			=> romPi2_output
    );
	
	delay_Pi2VariableDelay : ShiftRegister
	GENERIC MAP(
		size 				=> DELAY_PERMUTATION_2 - 1,
		width 				=> DEGREE_WIDTH
	)
	PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		input				=> in_variableDelay,
		output				=> romPi2_variableDelay
	);
	
	delay_Pi2ControlSignals : ShiftRegister
	GENERIC MAP(
		size 				=> DELAY_PERMUTATION_2 - 2,
		width 				=> ROM_ADDRESS_WIDTH + 1
	)
	PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		input(ROM_ADDRESS_WIDTH downto 1)	=> in_addressRom,
		input(0)							=> in_enableRom,
		output(ROM_ADDRESS_WIDTH downto 1)	=> romPi2_address,
		output(0)							=> romPi2_readEnable
	);
	
	delay_Pi0VariableDelay : ShiftRegister
	GENERIC MAP(
		size 				=> DURATION_PERMUTATION,
		width 				=> DEGREE_WIDTH
	)
	PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		input				=> romPi2_variableDelay,
		output				=> romPi0_variableDelay
	);
	
	delay_Pi0Control : ShiftRegister
	GENERIC MAP(
		size 				=> DURATION_PERMUTATION + 1,
		width 				=> ROM_ADDRESS_WIDTH + 1
	)
	PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		input(ROM_ADDRESS_WIDTH downto 1)	=> romPi2_address,
		input(0)							=> romPi2_readEnable,
		output(ROM_ADDRESS_WIDTH downto 1)	=> romPi0_address,
		output(0)							=> romPi0_readEnable
	);
	
	delay_Pi3Indices : VariableShiftRegister
	GENERIC MAP(
		size 				=> MAX_DEGREE,
		width 				=> ROM_PERMUTATION_WIDTH,
		offset				=> DELAY_PERMUTATION_3
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		active_size 		=> romPi2_variableDelay,
        input     			=> permutationIndicesPi2,
        output    			=> permutationIndicesPi3
    );
	
	delay_Pi1Indices : VariableShiftRegister
	GENERIC MAP(
		size 				=> MAX_DEGREE,
		width 				=> ROM_PERMUTATION_WIDTH,
		offset				=> DELAY_PERMUTATION_1 - 1
	)
    PORT MAP(
		clock				=> clock,
		enable				=> enable,
		reset				=> reset,
		active_size 		=> romPi0_variableDelay,
        input     			=> permutationIndicesPi0,
        output    			=> permutationIndicesPi1
    );
	
	-- Permutation of Indices
	InvertPi1Indices : process(permutationIndicesPi1)
	TYPE index_array is ARRAY (0 to PARALLELISM - 1) of std_logic_vector(PARALLELISM_WIDTH - 1 DOWNTO 0);
	variable temp_input		: index_array;
	variable index			: integer range 0 to PARALLELISM - 1;
	BEGIN
		for i in 0 to PARALLELISM - 1 loop
			temp_input(i) 	:= permutationIndicesPi1((PARALLELISM - i) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - i - 1) * PARALLELISM_WIDTH);
		end loop;
		
		permutationIndicesPi1_inverted	<= permutationIndicesPi1;
		for i in 0 to PARALLELISM - 1 loop
			index := to_integer(unsigned(temp_input(i)));
			permutationIndicesPi1_inverted((PARALLELISM - index) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - index - 1) * PARALLELISM_WIDTH) <= std_logic_vector(to_unsigned(i, PARALLELISM_WIDTH));
		end loop;
	END process;
	
	InvertPi2Indices : process(romPi2_output, in_oddIteration)
	TYPE index_array is ARRAY (0 to PARALLELISM - 1) of std_logic_vector(PARALLELISM_WIDTH - 1 DOWNTO 0);
	variable temp_input		: index_array;
	variable temp_output	: index_array;
	variable index			: integer range 0 to PARALLELISM - 1;
	BEGIN
		romPi2_output_inverted	<= romPi2_output;
		if in_oddIteration = '1' then
			for i in 0 to PARALLELISM - 1 loop
				temp_input(i) 	:= romPi2_output((PARALLELISM - i) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - i - 1) * PARALLELISM_WIDTH);
				temp_output(i)	:= (OTHERS => '0');
			end loop;
			
			for i in 0 to PARALLELISM - 1 loop
				index := to_integer(unsigned(temp_input(i)));
				temp_output(index) := std_logic_vector(to_unsigned(i, PARALLELISM_WIDTH));
			end loop;
			
			for i in 0 to PARALLELISM - 1 loop
				romPi2_output_inverted((PARALLELISM - i) * PARALLELISM_WIDTH - 1 downto (PARALLELISM - i - 1) * PARALLELISM_WIDTH) <= temp_output(i);
			end loop;
		end if;
	END process;
	
	
	-- out_permutationIndices_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 3 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi0;
	-- out_permutationIndices_vector(3 * ROM_PERMUTATION_WIDTH - 1 downto 2 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi1_output;
	-- out_permutationIndices_vector(2 * ROM_PERMUTATION_WIDTH - 1 downto 1 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi2;
	-- out_permutationIndices_vector(1 * ROM_PERMUTATION_WIDTH - 1 downto 0 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi3;
	OutputRegister : process(clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				permutationIndicesPi1_output	<= (OTHERS => '0');
				permutationIndicesPi2 			<= (OTHERS => '0');
				out_permutationIndices_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 3 * ROM_PERMUTATION_WIDTH)	<= (OTHERS => '0');
				out_permutationIndices_vector(3 * ROM_PERMUTATION_WIDTH - 1 downto 2 * ROM_PERMUTATION_WIDTH)	<= (OTHERS => '0');
				out_permutationIndices_vector(2 * ROM_PERMUTATION_WIDTH - 1 downto 1 * ROM_PERMUTATION_WIDTH)	<= (OTHERS => '0');
				out_permutationIndices_vector(1 * ROM_PERMUTATION_WIDTH - 1 downto 0 * ROM_PERMUTATION_WIDTH)	<= (OTHERS => '0');
			elsif enable = '1' then
				permutationIndicesPi1_output 	<= permutationIndicesPi1_inverted;
				permutationIndicesPi2 			<= romPi2_output_inverted;
				out_permutationIndices_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 3 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi0;
				out_permutationIndices_vector(3 * ROM_PERMUTATION_WIDTH - 1 downto 2 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi1_output;
				out_permutationIndices_vector(2 * ROM_PERMUTATION_WIDTH - 1 downto 1 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi2;
				out_permutationIndices_vector(1 * ROM_PERMUTATION_WIDTH - 1 downto 0 * ROM_PERMUTATION_WIDTH)	<= permutationIndicesPi3;
			end if;
		end if;
	END PROCESS;
	
END RTL;