-- Filename    	: DecodeScheduling.vhd
-- Author:	   	: Andreas Torno
-- Description 	: This component creates control information of decoding schedule e.g. rom memory access 
--				: and control signals for check nodes

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.widths.all;
USE work.cycles.all;

ENTITY DecodeScheduling IS
	GENERIC (
		widthIteration	: integer := 9;
		maxIteration	: integer := 1
	);
    PORT (
		-- Global control ports
		clock						: IN std_logic;
		enable						: IN std_logic;
		reset						: IN std_logic;
		-- Specific control ports
		out_rom_index				: OUT std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
		out_rom_enable				: OUT std_logic;
		out_first_value				: OUT std_logic;
		out_cn_index				: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_cn_degree				: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_iteration				: OUT std_logic_vector(widthIteration - 1 downto 0);
		out_lastIterationReached	: OUT std_logic
    );
END DecodeScheduling;

ARCHITECTURE RTL OF DecodeScheduling IS
	constant iterationDelay		: integer := 2 * (DEGREES(0) + 1) + 20;
	constant iterationDelayWidth: integer := 8;

	signal outMEM_degree		: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal outMEM_degree2		: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal outMEM_cycle			: std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
	signal sc_degree			: std_logic;
	signal c_degree				: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal ec_nextDegrees		: std_logic;
	signal c_nextDegree			: std_logic_vector(DEGREE_NUMBER_WIDTH - 1 downto 0);
	signal ec_cycle				: std_logic;
	signal enable2				: std_logic;
	signal c_cycle				: std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
	signal ec_iteration			: std_logic;
	signal ec_iteration2		: std_logic;
	signal c_iteration			: std_logic_vector(widthIteration - 1 downto 0);
	signal iteration			: std_logic_vector(widthIteration - 1 downto 0);
	signal eMEM_degrees			: std_logic;
	signal ec_iterationDelay	: std_logic;
	signal ec_iterationDelay2	: std_logic;
	signal ec_iterationDelay3	: std_logic;
	signal sc_iterationDelay	: std_logic;
	signal c_iterationDelay		: std_logic_vector(iterationDelayWidth - 1 downto 0);
	signal firstValue			: std_logic;
	signal lastIterationReached	: std_logic;
	signal rom_enable			: std_logic;
BEGIN
	sc_degree <= '1' when unsigned(c_degree) = 0 else '0';
	CounterDegree : process (clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				c_degree <= (OTHERS => '0');
			elsif enable = '1' and ec_iterationDelay2 = '0' and lastIterationReached = '0' then
				if sc_degree = '1' then
					c_degree <= outMEM_degree;
				else
					c_degree <= std_logic_vector(unsigned(c_degree) - 1);
				end if;
			end if;
		end if;
	END process;
	firstValue	<= '1' when unsigned(outMEM_degree) = unsigned(c_degree) else '0';
	
	ec_cycle 	<= '0' when unsigned(outMEM_degree) < unsigned(c_degree) or enable = '0' or ec_iterationDelay3 = '1' else '1';
	CounterCycle : process (clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				c_cycle <= (OTHERS => '1');
			elsif ec_cycle = '1' then
				if c_iteration(0) = '0' then
					c_cycle <= std_logic_vector(unsigned(c_cycle) + 1);
				else
					c_cycle <= std_logic_vector(unsigned(c_cycle) - 1);
				end if;
			end if;
		end if;
	END process;
	
	ec_nextDegrees <= '1' when unsigned(c_cycle) = unsigned(outMEM_cycle) and enable='1' and sc_iterationDelay = '0' else '0';
	CounterNextDegree : process (clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				c_nextDegree <= (OTHERS => '0');
			elsif ec_nextDegrees = '1' then
				if c_iteration(0) = '0' then
					c_nextDegree <= std_logic_vector(unsigned(c_nextDegree) + 1);
				else
					c_nextDegree <= std_logic_vector(unsigned(c_nextDegree) - 1);
				end if;
			end if;
		end if;
	END process;
	
	DegreesAccess : process (clock)
	--constant delay : integer := 2;
	BEGIN
		if clock'event and clock='1' then
			eMEM_degrees <= ec_nextDegrees or ec_iteration;
			if reset = '1' then
				eMEM_degrees <= '1';
				outMEM_degree <= std_logic_vector(to_unsigned(DEGREES(0), DEGREE_WIDTH));
				outMEM_cycle <= std_logic_vector(to_unsigned(CYCLES_FIRST(0), ROM_ADDRESS_WIDTH));-- - to_unsigned(delay, ROM_ADDRESS_WIDTH));
			elsif eMEM_degrees = '1' and ec_iterationDelay = '0' then
				if c_iteration(0) = '0' then
					outMEM_degree <= std_logic_vector(to_unsigned(DEGREES(to_integer(unsigned(c_nextDegree))), DEGREE_WIDTH));
					outMEM_cycle <= std_logic_vector(to_unsigned(CYCLES_FIRST(to_integer(unsigned(c_nextDegree))), ROM_ADDRESS_WIDTH));-- - to_unsigned(delay, ROM_ADDRESS_WIDTH));
				else
					outMEM_degree <= std_logic_vector(to_unsigned(DEGREES(to_integer(unsigned(c_nextDegree))), DEGREE_WIDTH));
					outMEM_cycle <= std_logic_vector(to_unsigned(CYCLES_SECOND(to_integer(unsigned(c_nextDegree))), ROM_ADDRESS_WIDTH));-- + to_unsigned(delay, ROM_ADDRESS_WIDTH));
				end if;
			end if;
		end if;
	END process;
	
	SetIterationDelay : process(c_cycle, c_iteration)
	--constant delay : integer := 2;
	variable temp : unsigned(ROM_ADDRESS_WIDTH - 1 downto 0);
	BEGIN
		sc_iterationDelay <= '0';
		if c_iteration(0) = '0' then
			temp := to_unsigned(CYCLES_FIRST(DEGREE_NUMBER - 1), ROM_ADDRESS_WIDTH);-- - delay;
			if unsigned(c_cycle) = temp then 
				sc_iterationDelay <= '1';
			end if;
		elsif c_iteration(0) = '1' then
			temp := to_unsigned(CYCLES_SECOND(0), ROM_ADDRESS_WIDTH);-- + delay;
			if unsigned(c_cycle) = temp then 
				sc_iterationDelay <= '1';
			end if;
		end if;
	END process;
	
	ec_iterationDelay <= '0' when unsigned(c_iterationDelay) = 0 else '1';
	CounterIterationDelay : process (clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				c_iterationDelay <= (OTHERS => '0');
			elsif sc_iterationDelay = '1' then
				c_iterationDelay <= std_logic_vector(to_unsigned(iterationDelay, iterationDelayWidth));
			elsif ec_iterationDelay = '1' then
				c_iterationDelay <= std_logic_vector(unsigned(c_iterationDelay) - 1);
			end if;
		end if;
	END process;
	
	DelayBetweenIterations : process (clock)
	--variable ec_iterationDelay1 : std_logic;
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' or ec_iteration = '1' then
				ec_iterationDelay2 <= '0';
				ec_iterationDelay3 <= '0';
			elsif enable = '1' then
				ec_iterationDelay3 <= ec_iterationDelay2;
				ec_iterationDelay2 <= ec_iterationDelay;
			end if;
		end if;
	END process;
	
	ec_iteration <= '1' when unsigned(c_iterationDelay) = 1 else '0';
	lastIterationReached <= '1' when unsigned(c_iteration) = maxIteration else '0';
	CounterIteration : process (clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				c_iteration <= (OTHERS => '0');
			elsif ec_iteration = '1' then
				c_iteration <= std_logic_vector(unsigned(c_iteration) + 1);
			end if;
		end if;
	END process;
	
	rom_enable					<= ec_cycle and not ec_iteration2 and enable2 and not lastIterationReached;
	InputOutputPortsAndRegisters : process (clock)
	BEGIN
		if clock'event and clock='1' then
			if reset = '1' then
				out_rom_index 	<= (OTHERS => '0');
				out_rom_enable	<= '0';
				out_first_value	<= '0';
				out_cn_index	<= (OTHERS => '0');
				out_cn_degree	<= std_logic_vector(to_unsigned(CYCLES_FIRST(0), DEGREE_WIDTH));
				out_iteration	<= (OTHERS => '0');
				ec_iteration2	<= '0';
				enable2			<= '0';
				out_lastIterationReached <= '0';
				iteration		<= (OTHERS => '0');
			elsif enable = '1' then
				enable2			<= enable;
				ec_iteration2	<= ec_iteration;
				if rom_enable = '1' then
					out_rom_index 	<= c_cycle;
				else
					out_rom_index 	<= (OTHERS => '0');
				end if;
				out_rom_enable	<= rom_enable;
				out_first_value	<= firstValue;
				outMEM_degree2	<= outMEM_degree;
				out_cn_index	<= std_logic_vector(unsigned(outMEM_degree2) - unsigned(c_degree));
				out_cn_degree	<= outMEM_degree;
				iteration		<= c_iteration;
				out_iteration	<= iteration;
				out_lastIterationReached <= lastIterationReached;
			end if;
		end if;
	END process;
	
	
END RTL;