-- Filename    	: ControlCN.vhd
-- Author:	   	: Andreas Torno
-- Description 	: This component generates control signal for all CN-FUs

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.widths.all;
USE work.delays.all;

ENTITY ControlCN IS
    PORT (
		-- Global control ports
		clock				: IN  std_logic;
		enable				: IN  std_logic;
		reset				: IN  std_logic;
		-- Specific control ports
		in_firstValue		: IN  std_logic;
		in_oddIteration		: IN  std_logic;
		in_variableDelay	: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_degreeIndex		: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_valid			: IN  std_logic_vector(PARALLELISM - 1 downto 0);
		-- CN control ports
		out_firstValue		: OUT std_logic;
		out_lastValue		: OUT std_logic;
		out_degreeIndexIn	: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_degreeIndexOut	: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_variableDelay	: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_valid			: OUT std_logic;
		-- Extrinsic memory control ports
		out_memoryEnable	: OUT std_logic;
		out_memoryRead		: OUT std_logic;
		out_memoryAddress	: OUT std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0)
    );
END ControlCN;

ARCHITECTURE RTL OF ControlCN IS
	COMPONENT ShiftRegister
	GENERIC (
		size 				: integer;
		width 				: integer
	);
	PORT (
		clock				: IN  std_logic;
		enable				: IN  std_logic;
		reset				: IN  std_logic;
		input     			: IN  std_logic_vector(width - 1 downto 0);
		output    			: OUT std_logic_vector(width - 1 downto 0)
	);
	END COMPONENT;
	
	COMPONENT VariableShiftRegister
	GENERIC (
		size 				: integer;
		width 				: integer;
		offset				: integer
	);
    PORT (
		clock				: IN  std_logic;
		enable				: IN  std_logic;
		reset				: IN  std_logic;
		active_size 		: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
        input     			: IN  std_logic_vector(width - 1 downto 0);
        output    			: OUT std_logic_vector(width - 1 downto 0)
    );
	END COMPONENT;
	
	signal c_readAddress		: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
	signal c_n_readAddress		: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
	signal readAddress			: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
	signal c_writeAddress		: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
	signal c_n_writeAddress		: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
	signal writeAddress			: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
	
	signal memoryEnable			: std_logic;
	signal simultaneousRW		: std_logic;
	signal firstValue			: std_logic_vector(DELAY_CN + 3 downto 0);
	signal firstValue_out		: std_logic;
	signal lastValue			: std_logic;
	signal lastValue_temp		: std_logic_vector(0 downto 0);
	signal lastValue_delayed	: std_logic;
	signal memoryRead			: std_logic;
	signal variableDelay		: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal degreeIndex			: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal valid				: std_logic_vector(MAX_DEGREE + 2 * DURATION_PERMUTATION + 1 downto 0);
	signal memoryAddress		: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);
BEGIN

	delay_indexIn : ShiftRegister
	GENERIC MAP(
		size 		=> DELAY_CN,
		width 		=> DEGREE_WIDTH
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> in_degreeIndex,
        output		=> degreeIndex
    );
	delay_variableDelay : ShiftRegister
	GENERIC MAP(
		size 		=> DELAY_CN - 1,
		width 		=> DEGREE_WIDTH
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> in_variableDelay,
        output		=> variableDelay
    );
	
	delay_IndexOut : VariableShiftRegister
	GENERIC MAP(
		size 		=> MAX_DEGREE,
		width 		=> DEGREE_WIDTH,
		offset		=> DURATION_CN
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> degreeIndex,
        output		=> out_degreeIndexOut,
		active_size => variableDelay
    );
	
	delay_LastValue : VariableShiftRegister
	GENERIC MAP(
		size 		=> MAX_DEGREE,
		width 		=> 1,
		offset		=> 0
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		input		=> firstValue(0 downto 0),
		output		=> lastValue_temp(0 downto 0),
		active_size => variableDelay
    );
	lastValue			<= lastValue_temp(0);

	firstValue(DELAY_CN + 3) <= in_firstValue;
	out_memoryEnable 	<= memoryEnable;
	memoryRead			<= firstValue(3);
	out_memoryRead		<= memoryRead;
	firstValue_out		<= firstValue(0);
	out_firstValue		<= firstValue_out;
	out_lastValue		<= lastValue;
	out_degreeIndexIn	<= degreeIndex;
	out_variableDelay	<= variableDelay;
	out_memoryAddress	<= memoryAddress;
	out_valid			<= valid(0);
	
	FirstValueShift : process(clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				firstValue(DELAY_CN + 2 downto 0) <= (OTHERS => '0');
			elsif enable = '1' then
				firstValue(DELAY_CN + 2 downto 0) <= firstValue(DELAY_CN + 3 downto 1);
			end if;
		end if;
	END process;
	
	MemoryEnableDetermination : process(clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				memoryEnable 	<= '0';
				simultaneousRW 	<= '0';
			elsif enable = '1' then
				simultaneousRW <= firstValue(4) and lastValue;
				if simultaneousRW = '1' then
					memoryEnable <= '1';
				else
					memoryEnable <= firstValue(4) or lastValue;
				end if;
				
			end if;
		end if;
	END process;
	
	c_n_readAddress <= std_logic_vector(unsigned(c_readAddress) + 1) when in_oddIteration = '0' else std_logic_vector(unsigned(c_readAddress) - 1);
	ReadCounter : process(clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				c_readAddress <= (OTHERS => '1');
				readAddress <= (OTHERS => '1');
			elsif firstValue(4) = '1' then
				c_readAddress <= c_n_readAddress;
				if in_oddIteration = '0' then
					readAddress <= c_n_readAddress;
				elsif in_oddIteration = '1' then
					readAddress <= c_readAddress;
				end if;
			end if;
		end if;
	END process;
	
	c_n_writeAddress <= std_logic_vector(unsigned(c_writeAddress) + 1) when in_oddIteration = '0' else std_logic_vector(unsigned(c_writeAddress) - 1);
	WriteCounter : process(clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				c_writeAddress <= (OTHERS => '1');
				writeAddress <= (OTHERS => '1');
			elsif lastValue = '1' then
				c_writeAddress <= c_n_writeAddress;
				if in_oddIteration = '0' then
					writeAddress <= c_n_writeAddress;
				elsif in_oddIteration = '1' then
					writeAddress <= c_writeAddress;
				end if;
			end if;
		end if;
	END process;
	
	AddressOutput : process (memoryEnable, memoryRead, writeAddress, readAddress)
	BEGIN
		if memoryEnable = '1' and memoryRead = '0' then
			memoryAddress <= writeAddress;
		else
			memoryAddress <= readAddress;
		end if;
	END process;
	
	Validity : process(clock)
	variable errors_present : std_logic;
	constant all_valid		: std_logic_vector(PARALLELISM - 1 downto 0) := (OTHERS => '1');
	BEGIN
		if clock'event and clock = '1' then
			valid(MAX_DEGREE + 2 * DURATION_PERMUTATION	downto 0) <= valid(MAX_DEGREE + 2 * DURATION_PERMUTATION + 1 downto 1);
			valid(MAX_DEGREE + 2 * DURATION_PERMUTATION + 1) <= '0';
			if reset = '1' then
				lastValue_delayed <= '0';
				errors_present := '0';
			else
				lastValue_delayed <= lastValue;
				if in_oddIteration = '1' and lastValue_delayed = '1' then
					if errors_present = '0' and in_valid /= all_valid then
						errors_present := '1'; 
					end if;
					if unsigned(writeAddress) = 0 then
						if errors_present = '0' then
							valid(MAX_DEGREE + 2 * DURATION_PERMUTATION + 1) <= '1';
						end if;
						errors_present := '0';
					end if;
				end if;
			end if;
		end if;
	END process;
		
END RTL;