--
-- Implementation Of SymbolNode
-- by Andreas Torno
--   

LIBRARY ieee;
USE ieee.std_logic_1164.all;

LIBRARY work;
USE work.types.all;
USE work.widths.all;
USE work.delays.all;

ENTITY SymbolNode IS
	PORT (
		-- Global control ports
		enable, clock, reset: IN  std_logic;
		
		-- Decoding control ports
		in_initialization	: IN  std_logic;
		in_saveValues		: IN  std_logic;
		in_oddIteration		: IN  std_logic;
		in_firstExtrinsic	: IN  std_logic;
		in_lastExtrinsic	: IN  std_logic;
		
		-- Memory control ports
		in_readEnable		: IN  std_logic;
		in_writeEnable		: IN  std_logic;
		in_oldSD_address	: IN  std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
		in_newSD_address	: IN  std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
		
		-- Soft Decisions ports
		in_input 			: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_output 			: OUT std_logic_vector(WIDTH_SD - 1 downto 0);
		in_SD_read			: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_SD_read			: OUT std_logic_vector(WIDTH_SD - 1 downto 0);
		in_SD_write			: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_SD_write		: OUT std_logic_vector(WIDTH_SD - 1 downto 0)
	);
END;

ARCHITECTURE Beh OF SymbolNode IS
	-- type memory is array(0 to size - 1) of std_logic_vector(width_data - 1 downto 0);
	-- impure function WriteSDsToFile (RAM : in memory) return memory is
		-- FILE		SDFile 		: text is in RamFileName;
		-- variable 	RamFileLine : line;
		-- variable 	RAM 		: memory;
	-- begin
		-- for I in memory'range loop
			-- readline (RamFile, RamFileLine);
			-- read (RamFileLine, RAM(I));
		-- end loop;
		-- return RAM;
	-- end function;
	
	COMPONENT MemorySinglePort
		GENERIC (
			size 			: integer := SN_PER_FU;
			width_data 		: integer := WIDTH_SD;
			width_address	: integer := ROM_SN_WIDTH - 2
		);
		
		PORT (
			clock			: IN std_logic;
			enable			: IN std_logic;
			reset			: IN std_logic;
			rw_enable		: IN std_logic;
			address			: IN std_logic_vector(width_address - 1 downto 0);
			input     		: IN std_logic_vector(width_data - 1 downto 0);
			output    		: OUT std_logic_vector(width_data - 1 downto 0)
		);
	END COMPONENT;
	
	COMPONENT MemoryDualPort
		GENERIC (
			size 			: integer := SN_PER_FU;
			width_data 		: integer := WIDTH_SD;
			width_address	: integer := ROM_SN_WIDTH - 2
		);
		
		PORT (
			-- Control ports
			clock			: IN std_logic;
			enable			: IN std_logic;
			reset			: IN std_logic;
			r_enable		: IN std_logic;
			w_enable		: IN std_logic;
			r_address		: IN std_logic_vector(width_address - 1 downto 0);
			w_address		: IN std_logic_vector(width_address - 1 downto 0);
			
			-- Data ports
			input     		: IN std_logic_vector(width_data - 1 downto 0);
			output    		: OUT std_logic_vector(width_data - 1 downto 0)
		);
	END COMPONENT;

	COMPONENT ShiftRegister
		GENERIC (
			size 			: integer := 1;
			width 			: integer := 32
		);
		
		PORT (
			-- Control ports
			clock			: IN std_logic;
			enable			: IN std_logic;
			reset			: IN std_logic;
			
			-- Data ports
			input     		: IN std_logic_vector(width-1 downto 0);
			output    		: OUT std_logic_vector(width-1 downto 0)
		);
	END COMPONENT;

	-- delayed input signals
	SIGNAL input						: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL oddIteration					: std_logic;
	SIGNAL saveValues					: std_logic;
	SIGNAL addIntrinsic					: std_logic;
	SIGNAL resetSum						: std_logic;
	
	-- Partial soft decision sum
	SIGNAL partialSum					: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL partialSum2					: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL toAddTo_inSD_Write			: std_logic_vector(WIDTH_SD - 1 downto 0);
	
	-- Memory Signals
	SIGNAL intrinsicMemory_address		: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	SIGNAL intrinsicMemory_output		: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL intrinsicMemory_readEnable	: std_logic;
	SIGNAL sdMemory1_output				: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL sdMemory1_input				: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL sdMemory1_readEnable			: std_logic;
	SIGNAL sdMemory1_writeEnable		: std_logic;
	SIGNAL sdMemory1_writeEnable_final	: std_logic;
	SIGNAL sdMemory1_readAddress		: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	SIGNAL sdMemory1_writeAddress		: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	SIGNAL sdMemory1_writeAddress_final	: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	
	SIGNAL sdMemory2_output				: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL sdMemory2_input				: std_logic_vector(WIDTH_SD - 1 downto 0);
	SIGNAL sdMemory2_readEnable			: std_logic;
	SIGNAL sdMemory2_writeEnable		: std_logic;
	SIGNAL sdMemory2_readEnable_final	: std_logic;
	SIGNAL sdMemory2_readAddress		: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	SIGNAL sdMemory2_writeAddress		: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	SIGNAL sdMemory2_readAddress_final	: std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
	
BEGIN
		Intrinsic_Memory : MemorySinglePort
	PORT MAP (
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		rw_enable	=> intrinsicMemory_readEnable,
		address		=> intrinsicMemory_address,
		input     	=> sdMemory1_input,
		output    	=> intrinsicMemory_output
	);
			
	-- BRAM_SINGLE_MACRO_inst : BRAM_SINGLE_MACRO
	-- generic map (
		-- BRAM_SIZE => "18Kb", -- Target BRAM, "18Kb" or "36Kb"
		-- DEVICE => "7SERIES", -- Target Device: "VIRTEX5", "7SERIES", "VIRTEX6, "SPARTAN6"
		-- DO_REG => 0, -- Optional output register (0 or 1)
		-- INIT => X"000000000", -- Initial values on output port
		-- INIT_FILE => "NONE",
		-- WRITE_WIDTH => WIDTH_SD, -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
		-- READ_WIDTH => WIDTH_SD, -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
		-- SRVAL => X"000000000", -- Set/Reset value for port output
		-- WRITE_MODE => "READ_FIRST" -- "WRITE_FIRST", "READ_FIRST" or "NO_CHANGE"
	-- )
	-- port map (
		-- DO => intrinsicMemory_output, -- Output data, width defined by READ_WIDTH parameter
		-- ADDR => intrinsicMemory_address, -- Input address, width defined by read/write port depth
		-- CLK => clock, -- 1-bit input clock
		-- DI => sdMemory1_input, -- Input data port, width defined by WRITE_WIDTH parameter
		-- EN => enable, -- 1-bit input RAM enable
		-- -- REGCE => REGCE, -- 1-bit input output register enable
		-- RST => reset, -- 1-bit input reset
		-- WE => not intrinsicMemory_readEnable -- Input write enable, width defined by write port depth
	-- );
	
	
	
	
		SD1_Memory : MemoryDualPort
	PORT MAP (
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		r_enable	=> sdMemory1_readEnable,
		w_enable	=> sdMemory1_writeEnable_final,
		r_address	=> sdMemory1_readAddress,
		w_address	=> sdMemory1_writeAddress_final,
		input     	=> sdMemory1_input,
		output    	=> sdMemory1_output
	);
		SD2_Memory : MemoryDualPort
	PORT MAP (
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		r_enable	=> sdMemory2_readEnable_final,
		w_enable	=> sdMemory2_writeEnable,
		r_address	=> sdMemory2_readAddress_final,
		w_address	=> sdMemory2_writeAddress,
		input     	=> sdMemory2_input,
		output    	=> sdMemory2_output
	);
	
		delay_DecodingControl : ShiftRegister
	GENERIC MAP(
		size 		=> 1 + DURATION_RAM_ACCESS,
		width 		=> 3
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input(0)	=> in_saveValues,
		input(1)	=> in_firstExtrinsic,
		input(2)	=> in_lastExtrinsic,
        output(0)	=> saveValues,
		output(1)	=> addIntrinsic,
		output(2)	=> resetSum
    );
		delay_PartialSum : ShiftRegister
	GENERIC MAP(
		size 		=> DURATION_PERMUTATION + 1,
		width 		=> WIDTH_SD
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> partialSum2,
        output		=> toAddTo_inSD_Write
    );	
		delay_SD1_RW : ShiftRegister
	GENERIC MAP(
		size 		=> DELAY_READ_WRITE - 1,
		width 		=> RAM_SN_ADDRESS_WIDTH + 1
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		input(RAM_SN_ADDRESS_WIDTH downto 1)	=> sdMemory1_readAddress,
		input(0)	=> sdMemory1_readEnable, 
        output(RAM_SN_ADDRESS_WIDTH downto 1)	=> sdMemory1_writeAddress,
        output(0)	=> sdMemory1_writeEnable
    );
		delay_SD2_RW : ShiftRegister
	GENERIC MAP(
		size 		=> DELAY_READ_WRITE,
		width 		=> RAM_SN_ADDRESS_WIDTH + 1
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		input(RAM_SN_ADDRESS_WIDTH downto 1)	=> sdMemory2_readAddress_final,
		input(0)	=> sdMemory2_readEnable_final,
        output(RAM_SN_ADDRESS_WIDTH downto 1)	=> sdMemory2_writeAddress,
        output(0)	=> sdMemory2_writeEnable
    );
	
	
	update: PROCESS (clock)
	BEGIN
		IF clock'event and clock = '1' THEN
			IF reset = '1' THEN
				out_output <= (OTHERS => '0');
				out_SD_read <= (OTHERS => '0');
				out_SD_write <= (OTHERS => '0');
			ELSIF enable = '1' THEN
				-- delays
				oddIteration <= in_oddIteration;
				input <= in_input;
				
				-- intrinsic memory access
				IF in_initialization = '1' THEN
					intrinsicMemory_address <= in_oldSD_address;
					intrinsicMemory_readEnable <= '0';
				ELSE
					intrinsicMemory_address <= in_newSD_address;
					intrinsicMemory_readEnable <= '1';
				END IF;
				
				-- SD1 memory access
				IF in_initialization = '1' THEN
					sdMemory1_writeAddress_final <= in_oldSD_address;
					sdMemory1_writeEnable_final <= in_initialization;
					sdMemory1_input <= in_input;
				ELSE
					sdMemory1_writeAddress_final <= sdMemory1_writeAddress;
					sdMemory1_writeEnable_final <= sdMemory1_writeEnable;
					IF oddIteration = '0' THEN
						sdMemory1_input <= in_SD_read;
						sdMemory1_readEnable <= in_readEnable;
						sdMemory1_readAddress <= in_oldSD_address;
					ELSE
						sdMemory1_input <= std_logic_vector(SD_TYPE(toAddTo_inSD_Write) + SD_TYPE(in_SD_write));
						sdMemory1_readEnable <= in_writeEnable;
						sdMemory1_readAddress <= in_newSD_address;
					END IF;
				END IF;
				
				-- SD2 memory access
				IF in_saveValues = '1' THEN
					sdMemory2_readAddress_final <= in_oldSD_address;
					sdMemory2_readEnable_final <= '1';
				ELSE
					sdMemory2_readAddress_final <= sdMemory2_readAddress;
					sdMemory2_readEnable_final <= sdMemory2_readEnable;
					IF oddIteration = '0' THEN	
						sdMemory2_input <= in_SD_write;
					ELSE
						sdMemory2_input <= in_SD_read;		
					END IF;
				END IF;
				
				
				-- Output Logic to Pi2 Network and output
				IF saveValues = '1' THEN
					out_output <= sdMemory2_output;
					out_SD_read <= (OTHERS => '0');
				ELSE
					IF oddIteration = '0' THEN	
						out_output <= sdMemory1_output;
						out_SD_read <= sdMemory1_output;
					ELSE
						out_output <= sdMemory2_output;
						out_SD_read <= sdMemory2_output;	
					END IF;
				END IF;
				
				-- Output Logic to Pi3 Network
				IF oddIteration = '0' THEN
					out_SD_write <= partialSum;
				ELSE
					IF resetSum = '0' THEN				
						out_SD_write <= sdMemory1_output;		
					ELSE
						out_SD_write <= (OTHERS => '0');
					END IF;
				END IF;			
			END IF;
		END IF;
	END PROCESS update;
	
	-- Value to add to Input of Pi3
	partialSum2 <= partialSum when addIntrinsic = '1' else input;
		
	partialSumCalculation: process (addIntrinsic, input, intrinsicMemory_output, sdMemory2_output)
	variable temp : SD_TYPE;
	BEGIN
		IF addIntrinsic = '1' THEN
			temp := SD_TYPE(intrinsicMemory_output);
		ELSIF addIntrinsic = '0' THEN
			temp := SD_TYPE(sdMemory2_output);
		END IF;
		partialSum <= std_logic_vector(SD_TYPE(input) + temp);
		-- IF addIntrinsic = '1' THEN
			-- partialSum <= std_logic_vector(SD_TYPE(input) + SD_TYPE(intrinsicMemory_output));
		-- ELSIF addIntrinsic = '0' THEN
			-- partialSum <= std_logic_vector(SD_TYPE(input) + SD_TYPE(sdMemory2_output));
		-- END IF;
	END PROCESS partialSumCalculation;
	
	SD2MemoryAccess: process (oddIteration, in_writeEnable, in_newSD_address, in_readEnable, in_oldSD_address)
	BEGIN
		IF oddIteration = '0' THEN
			sdMemory2_readEnable <= in_writeEnable;
			sdMemory2_readAddress <= in_newSD_address;
		ELSE
			sdMemory2_readEnable <= in_readEnable;
			sdMemory2_readAddress <= in_oldSD_address;
		END IF;
	END PROCESS SD2MemoryAccess;
END beh;
